# -*- coding: utf-8 -*-
'''
    Colmet
    ------
    Collecting metrics about jobs (task/process/cgroup/etc.) running in a
    distributed environnement

'''
__version__ = '0.5.2'
VERSION = __version__
